from __future__ import annotations
from typing import Any

from .publisher import Publisher


class Subject(Publisher):
    def send(self, value: Any) -> None:
        self.downstream(value)


class PassThroughSubject(Subject):
    def downstream(self, value: Any):
        self.upstream(value)



from __future__ import annotations

from typing import Optional

import discord
from discord import ui

from .view import View
from .provider import BaseProvider
from .message import Message


class ViewTracker(ui.View):
    def __init__(self, view: View, timeout: Optional[float] = 180.0):
        super().__init__(timeout=timeout)
        self.view: View = view
        self.items: dict[str, str] = {}
        self.body: Optional[Message] = None
        self.message: Optional[discord.Message] = None
        self.provider: Optional[BaseProvider] = None

    async def track(self, provider: BaseProvider):
        self.body = await self.view.body()
        while not isinstance(self.body, Message):
            self.body = await self.body.before_body()

        for item in self.body.get_discord_items():
            self.add_item(item)
        self.message = await provider.send_message(self.body._content, self.body._embeds, self)
        self.view._tracker = self
        self.provider = provider
        await self.view.on_appear()

    async def update(self):
        body = await self.view.body()
        while not isinstance(body, Message):
            body = await body.body()

        if self.body != body:
            self.body = body
            self.clear_items()
            for item in self.body.get_discord_items():
                self.add_item(item)
            await self.provider.edit_message(self.body._content, self.body._embeds, self)
            await self.view.on_update()

    async def _scheduled_task(self, item: ui.Item, interaction: discord.Interaction):
        self.provider.update_interaction(interaction)
        return await super(ViewTracker, self)._scheduled_task(item, interaction)



from __future__ import annotations
from typing import Optional

import aiohttp
from aiohttp import ClientSession

from .async_publisher import AsyncPublisher


class URLRequestPublisher(AsyncPublisher):
    def __init__(self, url: str, session: Optional[ClientSession] = None):
        super().__init__()
        self.url = url
        self.session = session or ClientSession()

    async def dispatch(self) -> None:
        async with self.session:
            async with self.session.get(self.url) as resp:
                await self.upstream(resp)

    def json(self, *args, **kwargs) -> URLRequestPublisher:
        async def _json(resp: aiohttp.ClientResponse) -> dict:
            return await resp.json(*args, **kwargs)
        self.map(_json)
        return self



# flake8: noqa
from .view import View
from .tracker import ViewTracker
from .provider import MessageProvider, InteractionProvider
from .button import LinkButton, Button
from .message import Message
from .observable_object import ObservableObject
from .state import state
from .published import published
from .select import SelectOption, Select
from .page import PaginationView, PaginationButtons, PageView
from .alert import Alert, ActionButton


__title__ = 'discord.ext.ui'
__author__ = 'sizumita'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020-present sizumita'
__version__ = "3.1.1"



from __future__ import annotations


from typing import Any, Callable, Optional


class Publisher:
    def __init__(self):
        self.child: Optional[Publisher] = None
        self.parent: Optional[Publisher] = None
        self.subscribers: list[Callable[[Any], None]] = []

    def set_child(self, child: Publisher) -> Publisher:
        self.child = child
        return self

    def chain(self, new_publisher: Publisher) -> Publisher:
        new_publisher.parent = self
        self.set_child(new_publisher)
        return new_publisher

    def downstream(self, value: Any):
        """
        親に伝播する
        :param value:
        :return:
        """
        if self.parent is not None:
            self.parent.downstream(value)

    def upstream(self, value: Any):
        """
        子供に伝播する
        :param value:
        :return:
        """
        if self.child is not None:
            self.child.upstream(value)

        for func in self.subscribers:
            func(value)

    def dispatch(self) -> Any:
        """
        subscriberが追加された時に初回起動させる
        :return:
        """
        if self.parent is not None:
            self.parent.dispatch()

    def sink(self, func: Callable[[Any], None]) -> Publisher:
        if self.child is not None:
            self.child.sink(func)
            return self
        self.subscribers.append(func)
        self.dispatch()
        return self

    def map(self, func: Callable[[Any], Any]) -> Publisher:
        if self.child is not None:
            self.child.map(func)
            return self
        new_publisher = MapPublisher(func)
        self.chain(new_publisher)
        return self


class MapPublisher(Publisher):
    def __init__(self, func: Callable[[Any], Any]) -> None:
        super().__init__()
        self.map_func = func

    def upstream(self, value: Any):
        if isinstance(value, list):
            value = [self.map_func(v) for v in value]
        else:
            value = self.map_func(value)

        for func in self.subscribers:
            func(value)



from __future__ import annotations
from typing import Callable, Optional, Type

import discord

from .message import Message
from .view import View
from .button import Button
from .state import state


class PaginationButtons:
    def first(self, now: int, last_page: int) -> Button:
        return Button("<<").style(discord.ButtonStyle.blurple).disabled(now == 0)

    def previous(self, now: int, last_page: int) -> Button:
        return Button("<").style(discord.ButtonStyle.green).disabled(now == 0)

    def indicator(self, now: int, last_page: int):
        return Button(f"{now+1}/{last_page+1}").style(discord.ButtonStyle.gray).disabled(True)

    def next(self, now: int, last_page: int) -> Button:
        return Button(">").style(discord.ButtonStyle.green).disabled(now == last_page)

    def last(self, now: int, last_page: int) -> Button:
        return Button(">>").style(discord.ButtonStyle.blurple).disabled(now == last_page)


class PaginationView(View):
    page = state("page")

    def __init__(
            self,
            views: list[PageView] | PageView,
            *,
            show_buttons: bool = True,
            show_disabled: bool = False,
            show_indicator: bool = True,
            check: Optional[Callable[[discord.Interaction], bool]] = None,
            first_page: int = 0,
            cls: Type[PaginationButtons] = PaginationButtons
    ):
        super(PaginationView, self).__init__()
        self._views = views
        self.show_buttons = show_buttons
        self.show_disabled = show_disabled,
        self.show_indicator = show_indicator
        self.check = check

        self.page = first_page  # index 0始まり
        self.max_page = len(self._views) - 1
        self.button_gen = cls()

    def change_page(self, interaction: discord.Interaction, page: int):
        if self.check is not None and not self.check(interaction):
            return
        self.page = page

    async def body(self) -> Message | View:
        view = self._views[self.page] if isinstance(self._views, list) else self._views
        buttons = []
        first = self.button_gen.first(self.page, self.max_page)
        if not (not self.show_disabled and getattr(first, "_disabled", False)):
            buttons.append(first.on_click(lambda x: self.change_page(x, 0)))

        prev = self.button_gen.previous(self.page, self.max_page)
        if not (not self.show_disabled and getattr(prev, "_disabled", False)):
            buttons.append(prev.on_click(lambda x: self.change_page(x, self.page - 1)))

        if self.show_indicator:
            indicator = self.button_gen.indicator(self.page, self.max_page)
            buttons.append(indicator)

        next_ = self.button_gen.next(self.page, self.max_page)
        if not (not self.show_disabled and getattr(next_, "_disabled", False)):
            buttons.append(next_.on_click(lambda x: self.change_page(x, self.page + 1)))

        last = self.button_gen.last(self.page, self.max_page)
        if not (not self.show_disabled and getattr(last, "_disabled", False)):
            buttons.append(last.on_click(lambda x: self.change_page(x, self.max_page)))

        await view.on_appear(self)
        body = await view.body(self)
        body.items([buttons])

        return body


class PageView(View):
    async def body(self, paginator: PaginationView) -> Message | View:
        return await super(PageView, self).body()

    async def on_appear(self, paginator: PaginationView) -> None:
        """
        ページが表示される前に実行されます。
        一つのPageViewだけをPaginationViewに登録していた場合でも、ページが変更されるごとに毎回実行されます。
        :param paginator:
        :return:
        """
        pass
